(**
   Conversion from BAPA (boolean algebras with Presburger arithmetic) sentences
   into PA (pure Presburger arithmetic) sentences.  
   Shows that BAPA is decidable and elementary.
   Viktor Kuncak, 16 July 2004.

   Updated with some QFBAPA util function in February 2007, vkuncak.
*)

(*
 * Copyright (C) 2004 Viktor Kuncak <vkuncak@mit.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *)

(* -------------------------------------------------- *)
(** Datatype of formulas *)

type ident = string

type binder = | Forallset | Existsset 
              | Forallint | Existsint 
              | Forallnat | Existsnat
	      | ForallProp | ExistsProp

type form = 
  | PropVar of ident
  | True
  | False
  | Not of form
  | And of form list
  | Or of form list
  | Impl of form * form
  | Bind of binder * ident * form
  | Less of intTerm * intTerm
  | Leq of intTerm * intTerm
  | Inteq of intTerm * intTerm
  | Seteq of setTerm * setTerm
  | Subseteq of setTerm * setTerm
and intTerm =
  | Intvar of ident
  | Const of int
  | Plus of intTerm list
  | Minus of intTerm * intTerm
  | Times of int * intTerm
  | IfThenElse of form * intTerm * intTerm
  | Div of intTerm * int
  | Card of setTerm
and setTerm =
  | Setvar of ident
  | Emptyset
  | Fullset
  | Complement of setTerm
  | Union of setTerm list
  | Inter of setTerm list

let maxcard = "MAXC"

(* -------------------------------------------------- *)
(**  Algorithm \alpha *)

  (* replace Seteq and Subseteq with Card(...)=0 *)
let simplify_set_relations (f:form) : form = 
  let rec sform f = match f with
  | PropVar _ -> f
  | False -> f
  | True -> f
  | Not f -> Not (sform f)
  | And fs -> And (List.map sform fs)
  | Or fs -> Or (List.map sform fs)
  | Impl(f1,f2) -> Impl(sform f1,sform f2)
  | Bind(b,id,f1) -> Bind(b,id,sform f1)
  | Less(it1,it2) -> Less(siterm it1, siterm it2)
  | Leq(it1,it2) -> Leq(siterm it1, siterm it2)
  | Inteq(it1,it2) -> Inteq(siterm it1,siterm it2)
  | Seteq(st1,st2) -> And[sform (Subseteq(st1,st2)); 
                          sform (Subseteq(st2,st1))]
  | Subseteq(st1,st2) -> Inteq(Card(Inter[st1;Complement st2]),Const 0)
  and siterm it = match it with
  | Intvar _ -> it
  | Const _ -> it
  | Plus its -> Plus (List.map siterm its)
  | Minus(it1,it2) -> Minus(siterm it1, siterm it2)
  | Times(k,it1) -> Times(k,siterm it1)
  | Div(it1,k) -> Div(siterm it1, k)
  | Card st -> it
  | IfThenElse(f1,it1,it2) -> 
      IfThenElse(sform f1, siterm it1, siterm it2)
  in sform f

  (* split f into quantifier sequence and body *)
let split_quants_body f = 
  let rec vl f acc = match f with
  | Bind(b,id,f1) -> vl f1 ((b,id)::acc)
  | f -> (acc,f)
  in vl f []

  (* extract set variables from quantifier sequence *)
let extract_set_vars quants = 
  List.map (fun (b,id) -> id)
    (List.filter (fun (b,id) -> (b=Forallset || b = Existsset))
       quants)

type partition = (ident * setTerm) list

  (* make canonical name for integer variable naming a cube *)
let make_name sts = 
  let rec mk sts = match sts with
  | [] -> ""
  | (Setvar _)::sts1 -> "1" ^ mk sts1
  | (Complement (Setvar _))::sts1 -> "0" ^ mk sts1
  | _ -> failwith "make_name: unexpected partition form"
  in "l_" ^ mk sts

  (* make all cubes over vs *)
let generate_partition (vs : ident list) : partition = 
  let add id ss = (Setvar id)::ss in
  let addc id ss = Complement (Setvar id)::ss in
  let add_set id inters =
    List.map (add id) inters @ 
    List.map (addc id) inters in 
  let mk_nm is = (make_name is, Inter is) in
  List.map mk_nm 
    (List.map List.rev
       (List.fold_right add_set vs [[]]))

  (* is the set term true in the set valuation
     -- reduces to propositional reasoning *)
let istrue (st:setTerm) (id,ivaluation) : bool = 
  let valuation = match ivaluation with
  | Inter v -> v
  | _ -> failwith "wrong valuation" in
  let lookup v = 
    if List.mem (Setvar v) valuation then true
    else if List.mem (Complement (Setvar v)) valuation then false
    else failwith ("istrue: unbound variable " ^ v ^ " in valuation") in
  let rec check st = match st with
  | Setvar v -> lookup v
  | Emptyset -> false
  | Fullset -> true
  | Complement st1 -> not (check st1)
  | Union sts -> List.fold_right (fun st1 t -> check st1 || t) sts false
  | Inter sts -> List.fold_right (fun st1 t -> check st1 && t) sts true
  in check st

  (* compute cardinality of set expression 
     as a sum of cardinalities of cubes *)
let get_sum (p:partition) (st:setTerm) : intTerm list = 
  List.map (fun (id,inter) -> Intvar id)
    (List.filter (istrue st) p)

  (* replace cardinalities of sets with sums of 
     variables denoting cube cardinalities *)
let replace_cards (p:partition) (f:form) : form = 
  let rec repl f = match f with
  | PropVar _ -> f
  | False -> False 
  | True -> True
  | Not f -> Not (repl f)
  | And fs -> And (List.map repl fs)
  | Or fs -> Or (List.map repl fs)
  | Impl(f1,f2) -> Impl(repl f1,repl f2)
  | Bind(b,id,f1) -> Bind(b,id,repl f1)
  | Less(it1,it2) -> Less(irepl it1,irepl it2)
  | Leq(it1,it2) -> Leq(irepl it1,irepl it2)
  | Inteq(it1,it2) -> Inteq(irepl it1,irepl it2)
  | Seteq(_,_)|Subseteq(_,_) -> failwith "failed inv in replace_cards"
  and irepl it = match it with
  | Intvar _ -> it
  | Const _ -> it
  | Plus its -> Plus (List.map irepl its)
  | Minus(it1,it2) -> Minus(irepl it1, irepl it2)
  | Times(k,it1) -> Times(k, irepl it1)
  | Div(it1, k) -> Div(irepl it1, k)
  | Card st -> Plus (get_sum p st)
  | IfThenElse(f1,it1,it2) -> IfThenElse(repl f1, irepl it1, irepl it2)
  in repl f

let apply_quants quants f =
  List.fold_right (fun (b,id) f -> Bind(b,id,f)) quants f

let make_defining_eqns id part = 
  let rec mk ps = match ps with
  | [] -> []
  | (id1,Inter (st1::sts1)) :: (id2,Inter (st2::sts2)) :: ps1
    when (st1=Setvar id && st2=Complement (Setvar id) && sts1=sts2) ->
      (Inter sts1,make_name sts1,id1,id2) :: mk ps1
  | _ -> failwith "make_triples: unexpected partition form" in
  let rename_last lss = match lss with
  | [(s,l,l1,l2)] -> [(s,maxcard,l1,l2)]
  | _ -> lss in
  rename_last (mk part)

  (* -------------------------- *)
  (* main loop of the algorithm *)
  (* -------------------------- *)
let rec eliminate_all quants part gr = match quants with
  | [] -> gr
  | (ExistsProp,id)::quants1 ->
      eliminate_all quants1 part (Bind(ExistsProp,id,gr))
  | (ForallProp,id)::quants1 ->
      eliminate_all quants1 part (Bind(ForallProp,id,gr))
  | (Forallint,id)::quants1 ->
      eliminate_all quants1 part (Bind(Forallint,id,gr))
  | (Existsint,id)::quants1 ->
      eliminate_all quants1 part (Bind(Existsint,id,gr))
  | (Existsnat,id)::quants1 ->
      eliminate_all quants1 part (Bind(Existsnat,id,gr))
  | (Forallnat,id)::quants1 ->
      eliminate_all quants1 part (Bind(Forallnat,id,gr))
  | (Existsset,id)::quants1 ->
      let eqns = make_defining_eqns id part in
      let newpart = List.map (fun (s,l',_,_) -> (l',s)) eqns in 
      let mk_conj (_,l',l1,l2) = Inteq(Intvar l',Plus[Intvar l1;Intvar l2]) in
      let conjs = List.map mk_conj eqns in
      let lquants = List.map (fun (l,_) -> (Existsnat,l)) part in
      let gr1 = apply_quants lquants (And (conjs @ [gr])) in
      eliminate_all quants1 newpart gr1
  | (Forallset,id)::quants1 ->
      let eqns = make_defining_eqns id part in
      let newpart = List.map (fun (s,l',_,_) -> (l',s)) eqns in 
      let mk_conj (_,l',l1,l2) = Inteq(Intvar l',Plus[Intvar l1;Intvar l2]) in
      let conjs = List.map mk_conj eqns in
      let lquants = List.map (fun (l,_) -> (Forallnat,l)) part in
      let gr1 = apply_quants lquants (Impl(And conjs, gr)) in
      eliminate_all quants1 newpart gr1

(* putting everything together *)

let alpha (f:form) : form =   
  (* assumes f in prenex form *)
  let (quants,fm) = split_quants_body f in
  let fm1 = simplify_set_relations fm in
  let setvars = List.rev (extract_set_vars quants) in
  let part = generate_partition setvars in
  let g1 = replace_cards part fm1 in
  eliminate_all quants part g1

(* ************************************************************
   Pretty printing 
 ************************************************************)

let string_of_form (f:form) : string = 
  let p s = "(" ^ s ^ ")" in
  let rec str f = match f with
  | PropVar id -> theid id
  | False -> "False"
  | True -> "True"
  | Not f1 -> "\\lnot " ^ str f1
  | And [] -> "\\btrue"
  | And fs -> ninfix "\\land" (List.map str fs)
  | Or [] -> "\\btrue"
  | Or fs -> ninfix "\\lor" (List.map str fs)
  | Impl(f1,f2) -> infix (str f1) "\\implies" (str f2)
  | Bind(Forallset,id,f1) -> binder "\\forallset" id f1
  | Bind(Existsset,id,f1) -> binder "\\existsset" id f1
  | Bind(Forallint,id,f1) -> binder "\\forallint" id f1
  | Bind(Existsint,id,f1) -> binder "\\existsint" id f1
  | Bind(Forallnat,id,f1) -> binder "\\forallnat" id f1
  | Bind(Existsnat,id,f1) -> binder "\\existsnat" id f1
  | Bind(ForallProp,id,f1) -> binder "\\forallprop" id f1
  | Bind(ExistsProp,id,f1) -> binder "\\existsprop" id f1
  | Less(it1,it2) -> infix (istr it1) "<" (istr it2)
  | Leq(it1,it2) -> infix (istr it1) "\\leq" (istr it2)
  | Inteq(it1,it2) -> infix (istr it1) "=" (istr it2)
  | Seteq(st1,st2) -> infix (sstr st1) "\\<seteq>" (sstr st2)
  | Subseteq(st1,st2) -> infix (sstr st1) "\\subseteq" (sstr st2)
  and istr it = match it with
  | Intvar id -> theid id
  | Const c -> theint c
  | Plus its -> ninfix "+" (List.map istr its)
  | Minus(it1,it2) -> infix (istr it1) "-" (istr it2)
  | Times(k,it) -> infix (theint k) "*" (istr it)
  | Div(it,k) -> infix (istr it) "div" (theint k)
  | Card st -> "|" ^ sstr st ^ "|"
  | IfThenElse(f1,it1,it2) ->
    p ("if " ^ p (str f1) ^ " then " ^ istr it1 ^ " else " ^ istr it2)
  and sstr st = match st with
  | Setvar id -> theid id
  | Emptyset -> "\\emptyset"
  | Fullset -> "\\Univ"
  | Complement st -> "({" ^ sstr st ^ "})^c"
  | Union sts -> ninfix "\\cup" (List.map sstr sts)
  | Inter sts -> ninfix "\\cap" (List.map sstr sts)
  and ninfix op ss = p (String.concat (" " ^ op ^ " ") ss)
  and infix s1 op s2 = p (s1 ^ " " ^ op ^ " " ^ s2)
  and binder b id f = b ^ " " ^ theid id ^ ". " ^ str f
  and theint k = Printf.sprintf "%d" k
  and theid s = 
    let sf s = if String.length s > 1 then "\\mathit{" ^ s ^ "}" else s in
    try 
      let i = String.index s '_' in
      sf (String.sub s 0 i) ^ "_{" ^
      String.sub s (i+1) (String.length s - i - 1) ^ "}"
    with Not_found -> sf s
  in str f

(* additional stuff *)

(** set variables in a formula *)
let set_vars_of (f : form) : ident list = 
  let rec fv (bv : ident list) (f : form) : ident list =
    match f with
      | PropVar _ -> []
      | False -> []
      | True -> []
      | Not f1 -> fv bv f1
      | And fs -> List.concat (List.map (fv bv) fs)
      | Or fs -> List.concat (List.map (fv bv) fs)
      | Impl(f1,f2) -> fv bv f1 @ fv bv f2
      | Bind(b,v,f1) -> fv (v::bv) f1
      | Less(it1,it2)
      | Leq(it1,it2)
      | Inteq(it1,it2) -> fvi bv it1 @ fvi bv it2
      | Seteq(st1,st2)
      | Subseteq(st1,st2) -> fvs bv st1 @ fvs bv st2
  and fvi bv it = match it with
    | Intvar _ -> []
    | Const c -> []
    | Plus its -> List.concat (List.map (fvi bv) its)
    | Minus(it1,it2) -> fvi bv it1 @ fvi bv it2
    | Times(k,it) -> fvi bv it
    | Div(it,k) -> fvi bv it
    | Card st -> fvs bv st
    | IfThenElse(f1,it1,it2) -> 
	fv bv f1 @ fvi bv it1 @ fvi bv it2
  and fvs bv st = match st with
    | Setvar id -> if List.mem id bv then [] else [id]
    | Emptyset -> []
    | Fullset -> []
    | Complement st -> fvs bv st
    | Union sts -> List.concat (List.map (fvs bv) sts)
    | Inter sts -> List.concat (List.map (fvs bv) sts)
  in
    Util.remove_dups (fv [] f)

let flip_binder (b : binder) : binder = match b with
  | Forallset -> Existsset
  | Existsset -> Forallset
  | Forallint -> Existsint
  | Existsint -> Forallint
  | Forallnat -> Existsnat
  | Existsnat -> Forallnat
  | ForallProp -> ExistsProp
  | ExistsProp -> ForallProp
let universal_binder (b : binder) : bool = match b with
  | ExistsProp
  | Existsset
  | Existsint
  | Existsnat -> false
  | ForallProp
  | Forallset
  | Forallint
  | Forallnat -> true

(** negation-normal form - ignoring formulas inside IfThenElse *)
let negation_normal_form (f0 : form) : form =
  let apply_pos (pos : bool) (f : form) : form =
    if pos then f else Not f in
  let rec nnf (pos : bool) (f : form) : form =
    match f with
      | PropVar _ 
      | True
      | False
      | Less(_,_)
      | Leq(_,_)
      | Inteq(_,_)
      | Seteq(_,_)
      | Subseteq(_,_) 
	  (* base cases *)
	-> apply_pos pos f
      | Not f1 -> nnf (not pos) f1
      | And fs -> 
	  if pos then And (List.map (nnf pos) fs)
	  else Or (List.map (nnf pos) fs)
      | Or fs -> 
	  if pos then Or (List.map (nnf pos) fs)
	  else And (List.map (nnf pos) fs)
      | Impl(f1,f2) -> nnf pos (Or [Not f1;f2])
      | Bind(b,id,f1) -> 
	  if pos then Bind(b,id,nnf pos f1)
	  else Bind(flip_binder b,id,nnf pos f1)
  in
    nnf true f0

exception NotQF of form
(** If f has only universal quantifiers, return a quantifier-free formula
    (since we are checking validity),
    otherwise return None *)
let get_quantifier_free (f : form) : form option =
  (* pre: f is in nnf *)
  let rec get_qf (f : form) : form =
    match f with
      | PropVar _ 
      | True
      | False
      | Less(_,_)
      | Leq(_,_)
      | Inteq(_,_)
      | Seteq(_,_)
      | Subseteq(_,_) -> 
	  f
      | Not f1 -> Not (get_qf f1)
      | And fs -> And (List.map get_qf fs)
      | Or fs -> Or (List.map get_qf fs)
      | Impl(f1,f2) -> get_qf (Or [Not f1;f2])
      | Bind(b,id,f1) when universal_binder b -> 
	  get_qf f1
      | Bind(_,_,f1) -> 
	  raise (NotQF f)
  in
    try
      Some (get_qf f)
    with NotQF bf ->
      Util.msg ("\nBAPA formula not quantifier-free due to\n" ^
		  string_of_form bf);
      None

let approximate_quantifier_free (f0 : form) : form =
  let rec approx (f : form) : form =
    match f with
      | PropVar _ 
      | True
      | False
      | Less(_,_)
      | Leq(_,_)
      | Inteq(_,_)
      | Seteq(_,_)
      | Subseteq(_,_) -> 
	  f
      | Not f1 -> Not (approx f1)
      | And fs -> And (List.map approx fs)
      | Or fs -> Or (List.map approx fs)
      | Impl(f1,f2) -> approx (Or [Not f1;f2])
      | Bind(b,id,f1) when universal_binder b -> 
	  approx f1
      | Bind(_,_,f1) -> False
  in
    approx f0

(** compute formula size *)
let formula_size (f : form) : int =
  let rec sum (xs : int list) = 
    match xs with 
      | [] -> 0
      | x::xs -> x + sum xs in
  let rec fsize (f : form) : int =
    match f with
      | PropVar _ | True | False -> 1
      | Not f -> 1 + fsize f
      | And fs | Or fs -> List.length fs + sum (List.map fsize fs)
      | Impl(f1,f2) -> 1 + fsize f1 + fsize f2
      | Bind(_,_,f1) -> 1 + fsize f1
      | Inteq(i1,i2) | Less(i1,i2) | Leq(i1,i2) -> 1 + isize i1 + isize i2
      | Seteq(s1,s2) | Subseteq(s1,s2) -> 1 + ssize s1 + ssize s2
  and isize (i : intTerm) : int =
    match i with
      | Intvar _ | Const _ -> 1
      | Plus is -> List.length is + sum (List.map isize is)
      | Minus(i1,i2) -> 1 + isize i1 + isize i2
      | Times(_,i1) | Div(i1,_) -> 1 + isize i1
      | IfThenElse(f1,i1,i2) -> 1 + fsize f1 + isize i1 + isize i2
      | Card s -> 1 + ssize s
  and ssize (s : setTerm) : int =
    match s with
      | Setvar _ | Emptyset | Fullset -> 1
      | Complement s1 -> 1 + ssize s
      | Union ss | Inter ss -> List.length ss + sum (List.map ssize ss)
  in
    fsize f

(* ****** examples ********************)

(*
let add_partition_def part fm = 
  let mk_conj (id,inter) = Inteq(Intvar id,Card(inter)) in
  let conjs = List.map mk_conj part in
  let body = And (conjs @ [fm]) in
  let mk_quant (id,_) f = Bind(Existsset,id,f) in
  List.fold_right mk_quant part body
*)

let process (f:form) = begin 
  print_string "Input formula is:\n\n$";
  print_string (string_of_form f);
  let res = alpha f in
  print_string "$\n\nOutput formula is:\n\n$";
  print_string (string_of_form res);
  print_string "$\n";
  res
end

let e=Setvar "e" and content=Setvar "content" and 
    content'=Setvar "content'" 
and size=Intvar "size" and size'=Intvar "size'"

let a = Setvar "A" and b = Setvar "B"

let example1 =
  Bind(Forallset,"e",
  Bind(Forallset,"content", Bind(Forallset,"content'",
  Bind(Forallint,"size", Bind(Forallint,"size'",
    Impl(And [Inteq(Card(e),Const 1);
              Inteq(Card(Inter[e;content]),Const 0);
              Inteq(Card(content),size);
              Seteq(content',Union[content;e]);
              Inteq(size',Plus[size;Const 1])],
         And[Less(Const 0,size'); Inteq(Card(content'),size')])
  )))))

let example2 =
  Bind(Existsset,"e",
  Bind(Existsset,"content", Bind(Existsset,"content'",
  Bind(Existsint,"size", Bind(Existsint,"size'",
    And [Inteq(Card(e),Const 1);
         Inteq(Card(Inter[e;content]),Const 0);
         Inteq(Card(content),size);
         Seteq(content',Union[content;e]);
         Inteq(size',Plus[size;Const 1]);
         Not (And[Less(Const 0,size'); Inteq(Card(content'),size')])]
  )))))

let example3 =
  Bind(Forallset,"A",
       Bind(Forallset, "B",
	    Impl(Subseteq(a,b),Seteq(a,b))))

let example4 =
  Bind(Forallset,"A",
       Bind(Forallset, "B",
	    Impl(And([Subseteq(b,a); Subseteq(a,b)]),Seteq(a,b))))

let example5 =
  Bind(Forallnat, "x",
  Bind(Forallset,"A",
       Bind(Forallset, "B",
            And [Less (Intvar "x", Const 10);
	         Impl(And([Subseteq(b,a); Subseteq(a,b)]),Seteq(a,b))])))

let example6 =
  Bind(Existsnat, "x",
  Bind(Forallset,"A",
       Bind(Forallset, "B",
            And [Less (Intvar "x", Const 10);
	         Impl(And([Subseteq(b,a); Subseteq(a,b)]),Seteq(a,b))])))

let example7 =
  Bind(Forallnat, "x",
  Bind(Forallset,"A",
       Bind(Forallset, "B",
	         Impl(And[Less (Intvar "x", Const 10);
                          Subseteq(b,a); Subseteq(a,b)],
                      And[Seteq(a,b); Less(Intvar "x", Const 9)]))))

let example8 =
  Bind(Forallset,"A",
       Bind(Forallset, "B",
            And[Leq(Card a, Card (Union [a;b]));
                Leq(Card (Union [a;b]), Plus[Card a; Card b])]))

(*
let pa_example = process example3
*)

