(** Decide the truth value of {!Form.form} formula
    that belongs to BAPA fragment.

    Convert {!Form.form} formula into bapa formula, then apply the algorithm {!Alpha} to convert it to Presburger arithmetic formula and then
    use decision procedure for Presburger arithmetic to decide the truth value of the formula.

   Currently using CVC Lite as PA decision procedure instead of Omega or LASH.
*)

open Common
open Form
open FormUtil
open TypeReconstruct
open TermRewrite
open Bapaform

(* Old prover invocation:

let lashIsUnsatisfiable (f : Form.form) : bool =
  let _ = print_string ("\ninput formula:\n" ^ (PrintForm.string_of_form f) ^ "\n"); flush_all () in
  let bf = Formbapa.formToBapa f in
  let _ = print_string ("\nBAPA:\n" ^ (Alpha.string_of_form bf) ^ "\n>>>\n"); flush_all () in
  let pf = Bapapres.bapaFormToPresForm (Alpha.alpha bf) in
    Presburger.lashIsUnsatisfiable pf

let valid (f : Form.form) : bool =
  let _ = print_string ("\ninput formula:\n" ^ (PrintForm.string_of_form f) ^ "\n"); flush_all () in
  let bf = Formbapa.formToBapa f in
  let _ = print_string ("\nBAPA:\n" ^ (Alpha.string_of_form bf) ^ "\n>>>\n"); flush_all () in
  let pf = Bapapres.bapaFormToPresForm (Alpha.alpha bf) in
    Presburger.omegaIsValid pf

let decide_sq0 (sqob : sq_obligation) (sqno:int) ~(options:options_t): bool =
  valid (form_of_sequent sqob.sqob_sq)
*)

(** The main function for deciding {!Form.form} formulas
    that belong to BAPA fragment. *)
let decide_sq (env0 : typeEnv) (sqob : sq_obligation) (sqno:int) 
    ~(options:options_t) : bool =
  let sq0 = sqob.sqob_sq in
  (* Eliminate free variables - substitute definition equality assumptions *)
  let s = elim_fv_in_seq false sq0 in
  (* Get types *)
  let fs = FormUtil.form_of_sequent s in
  let f = match fs with 
      TypedForm _ -> fs | _ -> TypedForm(fs, TypeApp(TypeBool, [])) in
  let env1 = get_annotated_types f in
  (* Resolve ambiguous operators and remove lambda abstraction *)
  let f0 = TypeReconstruct.disambiguate (unlambda f) in
  (* Rewrite set expressions and higher order constructs *)
  let rewrite_rules =
    [rewrite_function_eq;
     rewrite_pointsto;
     rewrite_FieldRead_FieldWrite]
  in
  let f1, env_useless = 
    try TermRewrite.rewrite true rewrite_rules env1 f0
    with _ -> f0, env1 in 
  let f2, env2 = reconstruct_types env_useless f1 in
  let f3 = remove_comments_and_typedform f2 in
  let bf = Formbapa.form_to_bapa false env2 f3 in
  let _ = Util.msg ("\nBAPA input:"); show_bf bf in
  let _ = flush_all() in
    (* invoke QFBAPA if possible, otherwise full BAPA *)
  let bf_nnf = Alpha.negation_normal_form bf in
  let _ = Util.msg("NNF:"); show_bf bf_nnf in
  let bf0 = Alpha.get_quantifier_free bf_nnf in
  let given_n = 
    try int_of_string (StringMap.find "sparse" options)
    with Not_found -> 0 
  in
  let _ = if given_n > 0 then Util.msg (Printf.sprintf "\n sparse=%d\n" given_n)
  in
  let abf = match bf0 with
    | Some bf1 -> 
	(Util.msg "\nObtained QFBAPA formula.";
	 if flag_positive options "fullbapa" then
	  (Util.msg "\nNevertheless, due to fullbapa option using full BAPA.\n";
	   Alpha.alpha bf)
	 else 
	   (Util.msg "Using QFBAPA procedure.\n";
	    Qfbapa.reduce_to_pa given_n bf1))
    | None -> 
	if flag_positive options "qfbapa" then
	  (Util.msg ("Due to 'qfbapa' command line option " ^
		       "approximating BAPA with QFBAPA.");
           Qfbapa.reduce_to_pa given_n (Alpha.approximate_quantifier_free bf_nnf))
	else 
	  (Util.msg "\nUsing full BAPA decision procedure.\n";
	   Alpha.alpha bf)
  in
  let _ = Util.msg(Printf.sprintf "\n ### Generated formula with %d nodes.\n" 
				     (Alpha.formula_size abf)) in
  let ab = bapa_to_form abf in
  let _ = Util.msg ("\nBAPA output in form:\n" ^ 
		       (PrintForm.string_of_form ab) ^ "\n") in
  let res_cvcl = SmtCvcl.cvcl_prove false ([], ab) options in
    res_cvcl
      (*
    if not res_cvcl then 
      Presburger.omegaIsValid (Bapapres.bapaFormToPresForm abf)
    else res_cvcl
      *)
