open Util
open Form
open FormUtil
open PrintForm
open BohneUtil
open Pred
open BohneProgram
open Region
open Ast
open Bf
open Bf_set
open Abstraction
open CachedDecider
open BfCachedDecider


(* debug messages *)
let debug_msg = Debug.debug_print_msg (Debug.register_debug_module "BohneRefine")

let guard_marker = "__guard"

let mark_as_guard f =
  match f with
  | App (Const Comment, [Const (StringConst s); f]) when s = guard_marker -> f
  |  _ -> mk_comment guard_marker f

exception RealCounterexample of string

(** Report counter-example *)
let report_counter_example is_real region trace last_wp orig_assn = 
  if is_real then print_endline "\n\nCOUNTER-EXAMPLE:\n"
  else print_endline "\n\nSPURIOUS COUNTER-EXAMPLE:\n";
  print_endline "Error trace:\n";
  let prettify_form f =
    let f0 = normalize_form false f in
    let f1 = strip_types f0 in
    string_of_form (unlambda f1)
  in
  let _ = List.fold_left (fun pp (region, gc, pp'_opt) ->
    print_abs_state stdout (Bf_set.join region.absstates);
    let pp' = safe_unsome (AstUtil.fresh_program_point ()) pp'_opt in
    print_newline ();
    print_command stdout pp (gc, pp'); pp') region.ppoint trace
  in
  Printf.printf "Violated property:\n\n%s\n\n" (prettify_form orig_assn);
  (if not is_real then 
       Printf.printf "Last violated weakest precondition:\n\n%s\n\n"
	 (prettify_form last_wp)
     else ()
    );
    raise (RealCounterexample "BohneCounterexample")

let is_guard_comment =
  let guard_regexp = Str.regexp (".*" ^ guard_marker) in
  fun c -> try Str.search_forward guard_regexp c 0 >= 0 with Not_found -> false

(** Rewrite rtrancl over updated fields *)
let rewrite_upd_rtrancl f =
  let rec find_update x y fs acc =
    match fs with
    | App (Const Eq, [TypedForm (App (TypedForm (App (Const FieldWrite, [fld; s; t]), _), [px]), _); py])::fs'
    | App (Const Eq, [App (TypedForm (App (Const FieldWrite, [fld; s; t]), _), [px]); py])::fs'
    | App (Const Eq, [py; TypedForm (App (TypedForm (App (Const FieldWrite, [fld; s; t]), _), [px]), _)])::fs'
      when equal px x && equal py y ->
	let f_fld = mk_eq (App (fld, [x]), y) in
	let s_neq_x = mk_neq (s, x) in
	Some (s, t, smk_and (f_fld :: s_neq_x :: List.rev_append fs' acc), smk_and (f_fld :: List.rev_append fs' acc))
    | App (Const And, more_fs)::fs' -> 
	find_update x y (List.rev_append more_fs fs') acc
    | f :: fs' -> find_update x y fs' (f :: acc)
    | [] -> None
  in
  let contains = function 
    | App (Const And, fs) -> 
	(fun f -> List.exists (equal f) fs)
    | _ -> 
	(fun _ -> false)
  in
  let rec r = function 
    | App (TypedForm (Var v, _), [Binder (Lambda, ([(x1, _); (x2, _)] as vs), f); x; y])
    | App (Var v, [Binder (Lambda, ([(x1, _); (x2, _)] as vs), f); x; y]) when v = str_rtrancl -> 
	( match find_update (mk_var x1) (mk_var x2) [f] [] with 
	| Some (s, t, restricted_f, new_f) ->
	    let rtcl x y = App (Var str_rtrancl, [Binder (Lambda, vs, new_f); x; y]) in
	    let restricted_rtcl x y = 
	      if equal x mk_null || contains restricted_f (mk_neq (x, mk_var x1)) then mk_eq (x, y) else
	      App (Var str_rtrancl, [Binder (Lambda, vs, restricted_f); x; y]) in
	    let new_rtcl = mk_or [restricted_rtcl x y; mk_and [rtcl x s; restricted_rtcl t y]] in
	    r new_rtcl
	| None -> App (Var v, [Binder (Lambda, vs, r f); x; y]) )
    | App (fn, args) -> 
	App (r fn, List.map r args)
    | TypedForm (f, ty) -> TypedForm (r f, ty)
    | Binder (b, vs, f) -> Binder (b, vs, r f)
    | f -> f
  in r f


(** Extracts predicates from a formula *)
let extract_preds (f : form) =
  let rec atoms acc = function
    | App (Const Not, fs)
    | App (Const Or, fs)
    | App (Const And, fs)
    | App (Const MetaAnd, fs)
    | App (Const Impl, fs) 
    | App (Const MetaImpl, fs)
    | App (Const Iff, fs) ->
	List.fold_left atoms acc fs
    | App (Const Comment, [Const StringConst c; f0]) ->
	if is_guard_comment c then acc
	else atoms acc f0
    | TypedForm (f0, _) -> atoms acc f0
    | App (Const Eq, [TypedForm (Var _, ty); TypedForm (Var _, _)]) 
    | App (Const Eq, [TypedForm (Const NullConst, ty); TypedForm (Var _, _)]) 
    | App (Const Eq, [TypedForm (Var _, ty); TypedForm (Const NullConst, _)]) 
    | App (Const Eq, [TypedForm (Const NullConst, ty); TypedForm (Const NullConst, _)]) 
      when ty = TypeApp(TypeObject, []) -> acc
    | f -> if Decider.test_valid ([], f) || Decider.test_valid ([f], mk_false) ||
      List.exists (equal f) acc then acc else f::acc
  in atoms [] f

(* simplify formula for predicate extraction *)
let simplify_form skolem_consts f0 =
  let f = concretize_preds f0 in
  let sq, _ = elim_free_vars false skolem_consts (sequent_of_form f) in
  let f1 = unlambda (form_of_sequent sq) in
  let f2, env = TypeReconstruct.reconstruct_types (get_annotated_types f1) f1 in
  let f3, env1 = 
    TermRewrite.rewrite true 
      [TermRewrite.rewrite_tree; 
       TermRewrite.rewrite_FieldRead_FieldWrite] 
      env f2
  in
  elim_quants f3, env1


let refine trace f0 skolem_consts_env =
  (* extract region from error trace which is going to be refined *)
  let region, _, _ = List.hd trace in
  (* remove type information from skolem constants *)
  let skolem_consts = List.map fst skolem_consts_env in
  let f, env = simplify_form skolem_consts f0 in
  let _ = debug_msg 1 (fun chan ->
    Printf.fprintf chan "\nRefining region for pp. %d with:\n" region.ppoint.pp_id;
    output_string chan (string_of_sequent (sequent_of_form f) ^ "\n"))
  in
  let region_join = Bf_set.join region.absstates in
  let context =
    (region_join, concretize_invariant region.preds)    
  in
  let inv = smk_and region.invariants in
  (* extract atomic formulae from f *)
  let atoms = extract_preds f in
  (* generate predicates from atoms *)
  let new_preds = List.fold_left (fun acc p ->
    match Util.intersect (fv p) skolem_consts with
    | [] -> 
	(* state predicate *)
	if decide context (smk_impl (inv, p)) ||
	decide context (smk_impl (inv, smk_not p))
	then acc else 
	  add_pred p [Inherit]::acc
    | [v] when List.assoc v env = mk_object_type ->
	(* unary predicate on objects *)
	let p_def = Binder (Comprehension, [(v, mk_object_type)], p) in
	union_preds [add_pred p_def [Inherit]] acc
    | _ -> 
	(* for now ignore all others *)
	acc) [] atoms
  in
  let _ = debug_msg 1 (fun chan -> 
    output_string chan "Adding predicates:\n";    
    List.iter (fun p -> print_pred_decl chan p) new_preds)
  in
  Some (region, new_preds)


(* try to factorize a conjunct from a disjunction of conjunctions to improve
   splitting into proof subgoals.
   e.g. (A & B | ~A & C | ~A | D)) becomes (A --> B) & (~A --> C | D) *)
let factorize disj =
  let rec conjuncts d =
    match d with
    | App (Const And, fs)
    | App (Const MetaAnd, fs) -> List.concat (List.rev_map conjuncts fs)
    | App (Const Not, [App (Const Or, fs)]) -> List.concat (List.rev_map (fun f -> conjuncts (smk_not f)) fs) 
    | _ -> [d]
  in
  match disj with
  | d::disj' ->
      (try find_map (fun f ->
	try let pos, neg = 
	  List.fold_left (fun (pos, neg) d ->
	    let cs, kind = List.fold_left 
		(fun (cs, kind) c -> 
		  if equal f c then (cs, Some true)
		  else if equal f (smk_not c) then (cs, Some false)
		  else (c::cs, kind)) ([], None) (conjuncts d)
	    in match kind with
	    | Some true -> (smk_and cs::pos, neg)
	    | Some false -> (pos, smk_and cs::neg)
	    | None -> raise Not_found) ([], []) disj
	in Some [smk_impl (f, smk_or pos); smk_impl (smk_not f, smk_or neg)]
	with Not_found -> None)
	  (conjuncts d)
      with Not_found -> [smk_or disj])
  | _ -> [smk_or disj]

(** Check assertions and analyze potential counter-examples *)
let check_assertions program region gc_assns =
  let check orig_assn =
    let rec analyze region trace pp_opt skolem_consts (src_states, gc, assn) =
      let mk_context abs_states =      
	let region_join = Bf_set.join abs_states in
	(region_join, concretize_invariant region.preds)
      in
      let inv = smk_and region.invariants in
      (* recursively split violated assertions to find a smallest violated subformula *)
      let rec check_assn src_states skolem_consts wp_assn =
	let context = mk_context src_states in
	let split_on_states cs assn =
	  if Bf_set.le (Bf_set.singleton (fst context)) src_states then Some (assn, cs)
	  else 
	    let _ = print_abs_states stdout src_states in
	    Bf_set.fold (function 
	    | None -> fun abs_state -> check_assn (Bf_set.singleton abs_state) cs assn
	    | res -> fun _ -> res) None src_states
	in
	if not (decide context (smk_impl (inv, wp_assn))) then
	  let split_assn = wp_assn in
	  let env = 
	    TypeReconstruct.get_env 
	      (form_of_sequent ([snd context (fst context); inv], split_assn)) 
	  in
	  let split_assn1, _ = 
	    TermRewrite.rewrite true 
	      [TermRewrite.rewrite_tree;
	       TermRewrite.rewrite_FieldRead_FieldWrite]
	      env (TypeReconstruct.disambiguate (unlambda split_assn))
	  in
	  let split_assn2 = Decider.generate_sequents [] (elim_quants split_assn1) in
	  let split_assns = 
	    List.map (fun ((asms, hyp), env) ->
	      let env' = env @ skolem_consts in
	      let hyp', _ =  TermRewrite.rewrite true [TermRewrite.rewrite_sets] env' hyp in
	      env' , form_of_sequent (asms, elim_quants hyp')) split_assn2 
	  in
	  try (match split_assns with
	  | [cs, f] ->
	      (match f with
	      | App (Const MetaImpl, [f;f1]) 
	      | App (Const Impl, [f;f1]) ->
		  (match f1 with 
		  | App(Const Or, disj)
		  | App(Const Comment, [_; App(Const Or, disj)]) ->
		      (match factorize disj with
		      | [f] -> split_on_states cs f
		      | assns' ->   
			  Some (Util.find_map (check_assn src_states cs) (List.rev_map (curry smk_impl f) assns')))
		  | App (Const Iff, [f11; f12])
		  | App (Const Comment, [_; App(Const Iff, [f11; f12])]) ->
		      Some (Util.find_map (uncurry (check_assn src_states)) 
			      [cs, smk_impl (smk_and [f; f11], f12);
			       cs, smk_impl (smk_and [f; f12], f11)])
		  | App (Const And, fs) 
		  | App (Const Comment, [_; App(Const And, fs)]) ->
		      Some (Util.find_map (check_assn src_states cs) (List.rev_map (curry smk_impl f) fs))
		  | App (Const Not, [App (Const Or, fs)])
		  | App (Const Comment, [_; App (Const Not, [App(Const Or, fs)])]) ->
		      Some (Util.find_map (check_assn src_states cs) (List.rev_map (fun f' -> smk_impl (f, smk_not f')) fs))
		  | App (Const Not, [App (Const And, f2::fs)])
		  | App (Const Comment, [_; App (Const Not, [App(Const And, f2::fs)])]) ->
		      check_assn src_states cs (smk_impl (smk_and [f;f2], smk_not (smk_and fs)))
 		  | f1 -> split_on_states cs (smk_impl (f, f1)))
	      | _ -> split_on_states cs f)
	  | _ -> Some (Util.find_map (uncurry (check_assn src_states)) split_assns))
	  with Not_found -> None
	else None
      in
      let trace' = (region, gc, pp_opt)::trace in
      let mk_wp f = smk_impl (mark_as_guard gc.gc_guard, program.wp gc.gc_command f) in
      let wp_assn = mk_wp assn in
      let _ = if pp_opt = None then 
	debug_msg 3 (fun chan -> 
	  output_string chan ("\nChecking assertion:\n" ^ (string_of_form (strip_types wp_assn)) ^ "\n"))
      in
      match check_assn src_states skolem_consts wp_assn with
      (* assertion is violated in current region *)
      | Some (wp_assn', skolem_consts') ->
	  let context = mk_context src_states in
	  debug_msg 3 (fun chan ->
	    output_string chan 
	    ("\nViolated formula:\n" ^ 
	     string_of_form (form_of_sequent ([snd context (fst context);inv], wp_assn')) ^ "\n"));
	  if is_root region then 
	    let f = smk_impl (smk_and [snd context (fst context);inv;program.initial_states], wp_assn') in
	    if not (decide trivial_context f)
	    then (* found real counter-example *)
	      report_counter_example true region trace' wp_assn' orig_assn
	    else (* full trace is spurious *)
	      refine trace' wp_assn' skolem_consts'
	  else (* continue analysis with parent region *)
	    let parent_gc, src_states, parent_region = region.parent in 
	    analyze !parent_region trace' (Some region.ppoint) skolem_consts' (src_states, parent_gc, wp_assn') 
      (* assertion is not violated in current region *)
      | None -> 
	  (match pp_opt with
	  | None -> 
            (* empty error trace, i.e. assertion verifies *) None
	  | Some pp -> 
	    (* spurious part of trace starts from current region *)
	      refine trace assn skolem_consts)
    in analyze
  in
  try Some (Util.find_map (fun (_, _, assn as gc_assn) -> check assn region [] None [] gc_assn) gc_assns)
  with Not_found -> None


(* count maximal number of quantified variables in the same scope *)
let count_max_bound =
  let add xs (x, _) = if List.mem x xs then xs else x::xs in
  let rec count bound old_max = function
    | App (Const Or, fs)
    | App (Const And, fs)
    | App (Const MetaAnd, fs)
    | App (Const Comment, Const StringConst _::fs)
    | App (Const Impl, fs) 
    | App (Const MetaImpl, fs)
    | App (Const Not, fs)
    | App (Const Iff, fs) ->
	List.fold_left (count bound) old_max fs
    | TypedForm (f1, _) -> count bound old_max f1
    | Binder (Forall, vs, f1)
    | Binder (Exists, vs, f1) ->
	count (List.fold_left add bound vs) old_max f1
    | _ -> max old_max (List.length bound)
  in count [] 0
  

(* check assertions for region [region] *)
let check program region =
  if not !CmdLine.refine then None else
  let loc = get_location program region.ppoint in
  let exit_loc = get_location program program.exit_point in
  let postconds gc1 = 
    List.map (fun (gc2, f) -> (region.absstates, compose program gc1 gc2, f))
      exit_loc.loc_assertions 
  in  
  let vcs = List.fold_left 
      (fun assertions (gc, succ_pp) ->
	if is_exit_pp program succ_pp && not !CmdLine.abstract_final
	then assertions @ postconds gc else assertions)
      (List.map (fun (gc, f) -> (region.absstates, gc, f)) loc.loc_assertions) loc.loc_cmds
  in
  let checked_region, vcs = 
    if not (is_exit_pp program region.ppoint) then
      region, vcs
    else
      let gc1, src_states, parent_ref = region.parent in
      let exit_invariant = concretize_invariant region.preds (Bf_set.join region.absstates) in
      let vcs = List.map (fun (gc2, f) -> 
	(src_states, compose program gc1 gc2, smk_impl (mark_as_guard exit_invariant, f))) exit_loc.loc_assertions
      in !parent_ref, vcs
  in
  let split_vcs0 = 
    List.concat (List.map (fun (r, c, f) ->
      List.map (fun f -> (r, c, f)) (split_conjuncts f)) vcs)
  in
   let split_vcs = 
   List.concat (List.map (fun (r, c, f) ->
      Bf_set.fold (fun acc abs_state -> (Bf_set.singleton abs_state, c, f)::acc) [] r) split_vcs0)
  in
  (* heuristic: sort properties increasingly wrt. the max. number of universally quantified variables in the same scope *)
  let sorted_vcs =
    let extended = List.rev_map (fun (_,_, f as a) ->
      (a, count_max_bound (fst (simplify_form [] f)))) split_vcs in
    let sorted = List.stable_sort (fun (_, c1) (_, c2) -> compare c1 c2) extended in
    List.map fst sorted
  in check_assertions program checked_region sorted_vcs
