(* Dispatch formulas to CVC Lite *)

open Form
open FormUtil
open Common
open SmtPrint
open SmtTranslate

(* This is a shorthand for printing debug messages for this module. *)
let debug_id : int = Debug.register_debug_module "SmtCvcl"
let debug_msg : (string -> unit) = Debug.debug_msg debug_id
let debug_is_on () : bool = Debug.debug_is_on debug_id

let default_opts () : options_t = 
  let smtPrint_defaults = [("TimeOut",       string_of_int !CmdLine.timeout);
			   ("MaxQuantInst",  "300");
			   ("TryHeuristics", "yes");
			   ("KeepRtrancl",   "no")
			  ] in
    map_of_list smtPrint_defaults

let heuristics_opts : options_t = 
  let smtPrint_defaults = [("TimeOut",       "20");
			   ("MaxQuantInst",  "300");
			   ("TryHeuristics", "yes");
			   ("KeepRtrancl",   "no")
			  ] in
    map_of_list smtPrint_defaults

let getOption (option_name : string) (options : options_t) : string = 
  StringMap.find option_name options

let cvcl_command o (in_fn : string) (out_fn : string) : string = 
  let timeout = getOption "TimeOut" o in
  let max_quant_inst = getOption "MaxQuantInst" o in
  let transclose = if flag_positive o "KeepRtrancl" then "+trans-closure " else "" in
  ("cvcl " ^ transclose ^ 
   " -max-quant-inst " ^ max_quant_inst ^ 
   " -timeout " ^ timeout ^ 
   " -lang smtlib " ^ in_fn ^ " > " ^ out_fn)

let infile_counter = (ref 0 : int ref)
let incr_infile () =
    infile_counter := !infile_counter + 1
let incr_infile_if_debug () =
  if (debug_is_on ()) then (incr_infile ())
let next_infile () : string = 
  "vc" ^ (string_of_int !infile_counter) ^ ".smt"

let outfile_counter = (ref 0 : int ref)
let incr_outfile () =
    outfile_counter := !outfile_counter + 1
let next_outfile () : string =
  "cvcl" ^ (string_of_int !outfile_counter) ^ ".out"

type cvcl_result = Proved | NotProved | TooManyQuantifiers

let check_cvcl_results (fn : string) : cvcl_result =
  let outfile = open_in fn in
    try
      let line = input_line outfile in
      let _ = close_in outfile in
      let valid_regexp = Str.regexp "Valid" in
      let unsat_regexp = Str.regexp "\\(Unsatisfiable\\)\\|\\(unsat\\)" in
      let invalid_regexp = Str.regexp "Not Valid" in
      let sat_regexp = Str.regexp "\\(Satisfiable\\)\\|\\(sat\\)" in
      let unknown_regexp = Str.regexp "Unknown" in
	if ((Str.string_match valid_regexp line 0) || 
	      (Str.string_match unsat_regexp line 0)) then
	  (debug_msg "Proved by cvcl.\n"; Proved)
	else
	  (if (Str.string_match unknown_regexp line 0) then
	     (debug_msg "There may be too many quantifiers.\n"; 
	      TooManyQuantifiers)
	   else
	     (if ((Str.string_match invalid_regexp line 0) ||
		    (Str.string_match sat_regexp line 0)) then
		debug_msg ("Disproved by cvcl.\n")
	      else
		debug_msg ("cvcl error.\n");
	      debug_msg ("Please see " ^ fn ^ " for complete output.\n");
	      NotProved))
    with End_of_file -> NotProved

let report_timeout (name : string) (o : options_t) : unit =
  Util.msg ("Failed to prove using " ^ name ^ " within the given timeout of " ^
	      (getOption "TimeOut" o) ^
	      "s.\nUse \'-timeout <seconds>\' to increase timeout.\n")

(* The suppress_errs flag was added for the -bohne option, which uses
   cvcl to prove loop invariants. We may want to specialize this
   further using a callback function for error messages, but for now,
   this will avoid the whole problem of the timeout errors that get
   spit out. *)
let invoke_cvcl (o : options_t)
    (s : sequent) (env : typeEnv) (fs : form list) (f : form) (suppress_errs : bool) : cvcl_result =
  let in_fn = (next_infile ()) in
  let fq_in_fn = Util.tmp_name in_fn in
  let fq_out_fn = Util.tmp_name (next_outfile ()) in
  let infile = open_out fq_in_fn in
  let smt_str = (convert_to_smt_string s in_fn env fs f) in
    output_string infile smt_str;
    close_out infile;
    try
      debug_msg ("Invoking cvcl on " ^ in_fn ^ ".\n");
      let exit_code = Util.run_with_timeout (cvcl_command o fq_in_fn fq_out_fn) !CmdLine.timeout in
	if (exit_code = 0) then
	  (check_cvcl_results fq_out_fn)
	else if suppress_errs then
	  NotProved
	else ((report_timeout "cvcl" o); NotProved)
    with Failure x -> Util.warn x; NotProved

let rec try_one_quantified_form (s : sequent) (env : typeEnv) (nqfs : form list) 
    (totry : form list) (f : form) (suppress_errs : bool) (o : options_t) : bool =
  match totry with
    | [] -> false
    | next :: rest ->
	let result = invoke_cvcl o s env (next :: nqfs) f suppress_errs in
	  match result with
	    | Proved -> true
	    | NotProved | TooManyQuantifiers ->
		(try_one_quantified_form s env nqfs rest f suppress_errs o)

let rec try_fewer_quantifiers (s : sequent) (env : typeEnv) (nqfs : form list) 
    (qfs0 : form list) (qfs1 : form list) (f : form) (suppress_errs : bool) 
    (o : options_t) : bool =
  match qfs0 with
    | [] -> false
    | next :: qfs -> 
	let result = invoke_cvcl o s env (nqfs @ qfs) f suppress_errs in
	  match result with
	    | Proved -> true
	    | NotProved | TooManyQuantifiers ->
		(try_fewer_quantifiers s env nqfs qfs (next :: qfs1) f suppress_errs o)

let try_heuristics (s : sequent) (env : typeEnv) (assumptions : form list) (goal : form) 
    (suppress_errs : bool) (o : options_t) : bool =
  let is_quantified (f : form) : bool =
    match f with
      | Binder(Forall, _, _)
      | Binder(Exists, _, _) -> true
      | _ -> false in
  let (qfs, nqfs) = List.partition is_quantified assumptions in
    ((try_fewer_quantifiers s env nqfs qfs [] goal suppress_errs o) ||
       (try_one_quantified_form s env nqfs qfs goal suppress_errs o))

let cvcl_prove (suppress_errs : bool) (s1 : sequent) (options : options_t) : bool =
  let s = (List.map remove_comments (fst s1), remove_comments (snd s1)) in
  let options = merge_opts (default_opts ()) options in
  try
    let (env, fs, f) = (process_sequent options s) in
    let result = invoke_cvcl options s env fs f suppress_errs in
      match result with
	| Proved -> incr_infile_if_debug (); true
	| NotProved -> incr_infile (); false
	| TooManyQuantifiers -> incr_infile ();
	    if flag_positive options "TryHeuristics" then
	      (try_heuristics s env fs f suppress_errs heuristics_opts)
	    else false
  with Failure x -> incr_infile (); Util.warn x; false
