(* strongest postcondition *)

open Ast
open Form
open FormUtil

exception Invariant_violation

(* This is a shorthand for printing debug messages for this module. *)
let debug_id : int = Debug.register_debug_module "Strongest_postcondition"
let debug_msg : (string -> unit) = Debug.debug_msg debug_id
let debug_is_on () : bool = Debug.debug_is_on debug_id

(* Generate a fresh variable. *)
let id_counter : int ref = ref 0
let increment_id (i : ident) : ident =
  id_counter := !id_counter + 1;
  (i ^ "_" ^ (string_of_int !id_counter))

(* Compute the strongest postcondition of f over (VarAssign ac) *)
let sp_VarAssign (ac : assign_command) (f : form) : form =
  let x = ac.assign_lhs in
  let e = ac.assign_rhs in
  let x0 = (increment_id x) in
  let v0 = mk_var x0 in (* give a new name to the old value of the variable *)
  let renameMap = [(x, v0)] in
  let f0 = subst renameMap f in (* substitute the new variable for the old in f *)
  let e0 = subst renameMap e in (* substitute the new variable for the old in e *)
  let f1 = smk_and [f0; mk_eq((mk_var x), e0)] in
    (* Implicitly existential quantified, needs to be on LHS of implication. *)
    FormUtil.smk_fixand_eq [x0] f1 (* get rid of some equalities *)
    
(* Compute the strongest postcondition of f over (Alloc ac) *)
let sp_Alloc (ac : alloc_command) (f : form) : form =
  let x = ac.alloc_lhs in
  let t = ac.alloc_type in
  let v = mk_var x in
  let x_in_alloc = mk_elem(v, all_alloc_objsf) in (* x is in the set of allocated objects *)
  let x_type = mk_elem(v, mk_var (obj_var t)) in (* x is in the set of objects of type t *)
  let x_nonnull = mk_neq(v, mk_null) in (* x ~= null *)
    (smk_and [f; x_in_alloc; x_type; x_nonnull])
 
(* Compute the strongest postcondition of f over (ArrayAlloc aac) *)
let sp_ArrayAlloc (aac : array_alloc_command) (f : form) : form =
  let x = aac.arr_alloc_lhs in
  let t = aac.arr_alloc_type in
  let dims = aac.arr_alloc_dims in
    match dims with
      | d0 :: drest ->
	  let v = mk_var x in
	  let x_in_alloc = mk_elem (v, all_alloc_objsf) in 
	  let x_has_type = mk_elem (v, mk_var (obj_var t)) in 
	  let x_nonnull = mk_neq (v, mk_null) in
	  let array_len = mk_eq ((mk_arrayLength v), d0) in
	    (smk_and [f; x_in_alloc; x_has_type; x_nonnull; array_len])
      | [] -> Util.warn ("LoopInvariantInference: array " ^ x ^ " has no dimensions.\n"); f

(* Compute the strongest postcondition of f over (Havoc hc) *)
let sp_Havoc (hc : havoc_command) (f : form) : form =
  let hr = hc.havoc_regions in
    if (List.length hr = 0) then f (* nothing to do here *)
    else
      let havoc_var (il : ident list) (to_havoc : form) : ident list =
	match to_havoc with
	  | Var x -> x :: il
	  | _ -> Util.warn ("LoopInvariantInference: don't know how to havoc " ^ 
			      (PrintForm.string_of_form to_havoc) ^ "\n"); il
      in
      let old_xs = List.fold_left havoc_var [] hr in (* collect identifiers *)
      let new_xs = List.map (fun (x) -> (increment_id x)) old_xs in (* get fresh names *)
      let new_vs = List.map (fun (x0) -> (mk_var x0)) new_xs in (* make new variables *)
      let renameMap = List.combine old_xs new_vs in
      let f0 = subst renameMap f in (* substitute the new variable for the old in f *)
      let til = List.map (fun (x) -> (x, TypeUniverse)) new_xs in
	smk_exists (til, f0)

(* Compute the strongest postcondition for a basic command. Since we
   are only using this for inference, we don't have to check asserts
   when we get to them.
*)
let sp_bc (bc : basic_command) (f : form) : form =
  match bc with
    | Skip
    | Hint _ -> f (* Hint is a special ast for Bohne. *)
    | VarAssign ac -> (sp_VarAssign ac f)
    | Alloc ac -> (sp_Alloc ac f)
    | ArrayAlloc aac -> (sp_ArrayAlloc aac f)
    | Assert {hannot_form = af; hannot_msg = am}
	(* We don't check asserts, but that's ok for inference. *)
    | NoteThat {hannot_form = af; hannot_msg = am}
	(* "NoteThat x" desugars into "Assert x; Assume x". Again, assert not checked. *)
    | Assume {annot_form = af; annot_msg = am} -> 
	smk_and [f; (mk_comment am af)]
    | Split {annot_form = af; annot_msg = am} ->
	(* "Split x" desugars into "Assert x; Havoc *; Assume x". Again, assert not checked. *)
	(mk_comment am af)
    | Havoc hc -> (sp_Havoc hc f)
    | ProcCall _ -> failwith "LoopInvariantInference: ProcCall should be desugared.\n"

(* Strongest postcondition with loops unrolled i times. Returns true
   if successful, false if a loop invariant is violated. *)
let sp_unrolled (f : form) (c : command) (i : int) : bool =
  let rec sp_unrolled_rec (f : form) (c : command) : form =
    match c with
      | Basic {bcell_cmd = bc} -> (sp_bc bc f)
      | Seq cl -> (List.fold_left sp_unrolled_rec f cl)
      | Choice cl -> smk_or (List.map (sp_unrolled_rec f) cl)
      | If {if_condition = ic; if_then = it; if_else = ie} ->
	  let tbranch = smk_and [ic; (sp_unrolled_rec f it)] in
	  let fbranch = smk_and [(smk_not ic); (sp_unrolled_rec f ie)] in
	    smk_or [tbranch; fbranch]
      | Loop lc -> 
	  begin
	    match lc.loop_inv with
	      | None -> 
		  se_loop_unroll lc f i
	      | Some loop_inv -> 
		  se_loop_inv lc f loop_inv
	  end
      | Return {return_exp = re} ->
	    match re with
	      | None -> f
	      | Some f0 -> smk_and [f; mk_eq (result_f, f0)]
  and se_loop_unroll (lc : loop_command) (f : form) (j : int) : form =
    let after_pre = sp_unrolled_rec f lc.loop_prebody in
      if (j = 0) then 
	smk_and [after_pre; (smk_not lc.loop_test)]
      else
	let after_test = smk_and [after_pre; lc.loop_test] in
	let after_post = sp_unrolled_rec after_test lc.loop_postbody in
	  (se_loop_unroll lc after_post (j - 1))
  and se_loop_inv (lc : loop_command) (f : form) (loop_inv : form) : form =
    let implies_inv = smk_impl (f, loop_inv) in
      if (not (Decider.ff_valid implies_inv)) then
	(debug_msg ("[not initially true]: " ^ 
		      (PrintForm.string_of_form implies_inv) ^ "\n");
	 raise Invariant_violation)
      else
	debug_msg ("[initially true]: " ^ 
			(PrintForm.string_of_form implies_inv) ^ "\n");
      let after_pre = sp_unrolled_rec loop_inv lc.loop_prebody in
      let after_test = smk_and [after_pre; lc.loop_test] in
      let after_post = sp_unrolled_rec after_test lc.loop_postbody in
      let inductive = smk_impl (after_post, loop_inv) in
	if (not (Decider.ff_valid inductive)) then
	  (debug_msg ("[not inductive]: " ^ 
			   (PrintForm.string_of_form inductive) ^ "\n");
	   raise Invariant_violation)
	else
	  debug_msg ("[inductive]: " ^ 
			  (PrintForm.string_of_form inductive) ^ "\n");
	smk_and [after_pre; (smk_not lc.loop_test)] in
    try
      let _ = (sp_unrolled_rec f c) in true
    with Invariant_violation -> false

